@@ -1,7 +1,8 @@
-const { resolve, basename } = require('path')
+const { join, resolve, basename } = require('path')
 const t = require('tap')
 const runScript = require('@npmcli/run-script')
 const localeCompare = require('@isaacs/string-locale-compare')('en')
+const tnock = require('../fixtures/tnock')
 
 // mock rimraf so we can make it fail in rollback tests
 const realRimraf = require('rimraf')
@@ -1770,6 +1771,8 @@ t.test('running lifecycle scripts of unchanged link nodes on reify', async t =>
 
   t.ok(fs.lstatSync(resolve(path, 'a/a-prepare')).isFile(),
     'should run prepare lifecycle scripts for links directly linked to the tree')
+  t.ok(fs.lstatSync(resolve(path, 'a/a-post-install')).isFile(),
+    'should run postinstall lifecycle scripts for links directly linked to the tree')
 })
 
 t.test('save-prod, with optional', async t => {
@@ -2465,6 +2468,62 @@ t.test('add local dep with existing dev + peer/optional', async t => {
   t.equal(tree.children.size, 1, 'children')
 })
 
+t.test('runs dependencies script if tree changes', async (t) => {
+  const path = t.testdir({
+    'package.json': JSON.stringify({
+      name: 'root',
+      version: '1.0.0',
+      dependencies: {
+        abbrev: '^1.1.1',
+      },
+      scripts: {
+        predependencies: `node -e "require('fs').writeFileSync('ran-predependencies', '')"`,
+        dependencies: `node -e "require('fs').writeFileSync('ran-dependencies', '')"`,
+        postdependencies: `node -e "require('fs').writeFileSync('ran-postdependencies', '')"`,
+      },
+    }),
+  })
+
+  await reify(path)
+
+  for (const script of ['predependencies', 'dependencies', 'postdependencies']) {
+    const expectedPath = join(path, `ran-${script}`)
+    t.ok(fs.existsSync(expectedPath), `ran ${script}`)
+    // delete the files after we assert they exist
+    fs.unlinkSync(expectedPath)
+  }
+
+  // reify again without changing dependencies
+  await reify(path)
+
+  for (const script of ['predependencies', 'dependencies', 'postdependencies']) {
+    const expectedPath = join(path, `ran-${script}`)
+    // and this time we assert that they do _not_ exist
+    t.not(fs.existsSync(expectedPath), `did not run ${script}`)
+  }
+
+  // take over console.log as run-script is going to print a banner for these because
+  // they're running in the foreground
+  const _log = console.log
+  t.teardown(() => {
+    console.log = _log
+  })
+  const logs = []
+  console.log = (msg) => logs.push(msg)
+  // reify again, this time adding a new dependency
+  await reify(path, { foregroundScripts: true, add: ['once@^1.4.0'] })
+  console.log = _log
+
+  t.match(logs, [/predependencies/, /dependencies/, /postdependencies/], 'logged banners')
+
+  // files should exist again
+  for (const script of ['predependencies', 'dependencies', 'postdependencies']) {
+    const expectedPath = join(path, `ran-${script}`)
+    t.ok(fs.existsSync(expectedPath), `ran ${script}`)
+    fs.unlinkSync(expectedPath)
+  }
+})
+
 t.test('save package.json on update', t => {
   t.test('should save many deps in multiple package.json when using save=true', async t => {
     const path = fixture(t, 'workspaces-need-update')
@@ -2622,5 +2681,375 @@ t.test('save package.json on update', t => {
       'should save no top level dep update to root package.json'
     )
   })
+
+  t.test('should not throw when trying to update a link dep', async t => {
+    const path = t.testdir({
+      one: {
+        'package.json': JSON.stringify({
+          name: 'one',
+          version: '1.0.0',
+          dependencies: {
+            two: 'file:../two',
+          },
+        }),
+      },
+      two: {
+        'package.json': JSON.stringify({
+          name: 'two',
+          version: '1.0.0',
+        }),
+      },
+    })
+
+    await t.resolves(reify(resolve(path, 'one'), { update: true, save: true, workspaces: [] }))
+
+    t.equal(
+      require(resolve(path, 'one', 'package.json')).dependencies.two,
+      'file:../two',
+      'should have made no changes'
+    )
+  })
   t.end()
 })
+
+t.test('installLinks', (t) => {
+  t.test('when true, packs and extracts instead of symlinks', async (t) => {
+    const path = t.testdir({
+      a: {
+        'package.json': JSON.stringify({
+          name: 'a',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            b: 'file:../b',
+          },
+        }),
+        'index.js': '',
+      },
+      b: {
+        'package.json': JSON.stringify({
+          name: 'b',
+          version: '1.0.0',
+          main: 'index.js',
+        }),
+        'index.js': '',
+      },
+    })
+
+    await reify(resolve(path, 'a'), { installLinks: true })
+
+    const installedB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(installedB.isDirectory(), 'a/node_modules/b is a directory')
+  })
+
+  t.test('when false, symlinks', async (t) => {
+    const path = t.testdir({
+      a: {
+        'package.json': JSON.stringify({
+          name: 'a',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            b: 'file:../b',
+          },
+        }),
+        'index.js': '',
+      },
+      b: {
+        'package.json': JSON.stringify({
+          name: 'b',
+          version: '1.0.0',
+          main: 'index.js',
+        }),
+        'index.js': '',
+      },
+    })
+
+    await reify(resolve(path, 'a'), { installLinks: false })
+
+    const installedB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(installedB.isSymbolicLink(), 'a/node_modules/b is a symlink')
+  })
+
+  t.test('when symlinks exist, installLinks set to true replaces them with dirs', async (t) => {
+    const path = t.testdir({
+      a: {
+        'package.json': JSON.stringify({
+          name: 'a',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            b: 'file:../b',
+          },
+        }),
+        'index.js': '',
+      },
+      b: {
+        'package.json': JSON.stringify({
+          name: 'b',
+          version: '1.0.0',
+          main: 'index.js',
+        }),
+        'index.js': '',
+      },
+    })
+
+    await reify(resolve(path, 'a'), { installLinks: false, save: true })
+
+    const firstB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(firstB.isSymbolicLink(), 'a/node_modules/b is a symlink')
+
+    await reify(resolve(path, 'a'), { installLinks: true, save: true })
+
+    const secondB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(secondB.isDirectory(), 'a/node_modules/b is now a directory')
+  })
+
+  t.test('when directories exist, installLinks set to false replaces them with symlinks', async (t) => {
+    const path = t.testdir({
+      a: {
+        'package.json': JSON.stringify({
+          name: 'a',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            b: 'file:../b',
+          },
+        }),
+        'index.js': '',
+      },
+      b: {
+        'package.json': JSON.stringify({
+          name: 'b',
+          version: '1.0.0',
+          main: 'index.js',
+        }),
+        'index.js': '',
+      },
+    })
+
+    await reify(resolve(path, 'a'), { installLinks: true })
+
+    const firstB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(firstB.isDirectory(), 'a/node_modules/b is a directory')
+
+    await reify(resolve(path, 'a'), { installLinks: false })
+
+    const secondB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(secondB.isSymbolicLink(), 'a/node_modules/b is now a symlink')
+  })
+
+  t.test('when installLinks is true, dependencies of links are installed', async (t) => {
+    const path = t.testdir({
+      a: {
+        'package.json': JSON.stringify({
+          name: 'a',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            b: 'file:../b',
+          },
+        }),
+        'index.js': '',
+      },
+      b: {
+        'package.json': JSON.stringify({
+          name: 'b',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            abbrev: '^1.0.0',
+          },
+        }),
+        'index.js': '',
+      },
+    })
+
+    await reify(resolve(path, 'a'), { installLinks: true })
+
+    const installedB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(installedB.isDirectory(), 'a/node_modules/b is a directory')
+
+    const abbrev = fs.lstatSync(resolve(path, 'a/node_modules/abbrev'))
+    t.ok(abbrev.isDirectory(), 'abbrev got installed')
+  })
+
+  t.test('workspaces are always symlinks, even with installLinks set to true', async (t) => {
+    const path = t.testdir({
+      a: {
+        'package.json': JSON.stringify({
+          name: 'a',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            b: 'file:../b',
+            c: '^1.0.0',
+          },
+          workspaces: ['./c'],
+        }),
+        'index.js': '',
+        c: {
+          'package.json': JSON.stringify({
+            name: 'c',
+            version: '1.0.0',
+            main: 'index.js',
+          }),
+          'index.js': '',
+        },
+      },
+      b: {
+        'package.json': JSON.stringify({
+          name: 'b',
+          version: '1.0.0',
+          main: 'index.js',
+          dependencies: {
+            abbrev: '^1.0.0',
+          },
+        }),
+        'index.js': '',
+      },
+    })
+
+    await reify(resolve(path, 'a'), { installLinks: true })
+
+    const installedB = fs.lstatSync(resolve(path, 'a/node_modules/b'))
+    t.ok(installedB.isDirectory(), 'a/node_modules/b is a directory')
+
+    const installedC = fs.lstatSync(resolve(path, 'a/node_modules/c'))
+    t.ok(installedC.isSymbolicLink(), 'a/node_modules/c is a symlink')
+
+    const abbrev = fs.lstatSync(resolve(path, 'a/node_modules/abbrev'))
+    t.ok(abbrev.isDirectory(), 'abbrev got installed')
+  })
+
+  t.end()
+})
+
+t.only('should preserve exact ranges, missing actual tree', async (t) => {
+  const Arborist = require('../../lib/index.js')
+  const abbrev = resolve(__dirname,
+    '../fixtures/registry-mocks/content/abbrev/-/abbrev-1.1.1.tgz')
+  const abbrevTGZ = fs.readFileSync(abbrev)
+
+  const abbrevPackument = JSON.stringify({
+    _id: 'abbrev',
+    _rev: 'lkjadflkjasdf',
+    name: 'abbrev',
+    'dist-tags': { latest: '1.1.1' },
+    versions: {
+      '1.1.1': {
+        name: 'abbrev',
+        version: '1.1.1',
+        dist: {
+          tarball: 'https://registry.npmjs.org/abbrev/-/abbrev-1.1.1.tgz',
+        },
+      },
+    },
+  })
+
+  const abbrevPackument2 = JSON.stringify({
+    _id: 'abbrev',
+    _rev: 'lkjadflkjasdf',
+    name: 'abbrev',
+    'dist-tags': { latest: '1.1.1' },
+    versions: {
+      '1.1.1': {
+        name: 'abbrev',
+        version: '1.1.1',
+        dist: {
+          tarball: 'https://registry.garbage.org/abbrev/-/abbrev-1.1.1.tgz',
+        },
+      },
+    },
+  })
+
+  t.only('host should not be replaced replaceRegistryHost=never', async (t) => {
+    const testdir = t.testdir({
+      project: {
+        'package.json': JSON.stringify({
+          name: 'myproject',
+          version: '1.0.0',
+          dependencies: {
+            abbrev: '1.1.1',
+          },
+        }),
+      },
+    })
+
+    tnock(t, 'https://registry.github.com')
+      .get('/abbrev')
+      .reply(200, abbrevPackument)
+
+    tnock(t, 'https://registry.npmjs.org')
+      .get('/abbrev/-/abbrev-1.1.1.tgz')
+      .reply(200, abbrevTGZ)
+
+    const arb = new Arborist({
+      path: resolve(testdir, 'project'),
+      registry: 'https://registry.github.com',
+      cache: resolve(testdir, 'cache'),
+      replaceRegistryHost: 'never',
+    })
+    await arb.reify()
+  })
+
+  t.only('host should be replaced replaceRegistryHost=npmjs', async (t) => {
+    const testdir = t.testdir({
+      project: {
+        'package.json': JSON.stringify({
+          name: 'myproject',
+          version: '1.0.0',
+          dependencies: {
+            abbrev: '1.1.1',
+          },
+        }),
+      },
+    })
+
+    tnock(t, 'https://registry.github.com')
+      .get('/abbrev')
+      .reply(200, abbrevPackument)
+
+    tnock(t, 'https://registry.github.com')
+      .get('/abbrev/-/abbrev-1.1.1.tgz')
+      .reply(200, abbrevTGZ)
+
+    const arb = new Arborist({
+      path: resolve(testdir, 'project'),
+      registry: 'https://registry.github.com',
+      cache: resolve(testdir, 'cache'),
+      replaceRegistryHost: 'npmjs',
+    })
+    await arb.reify()
+  })
+
+  t.only('host should be always replaceRegistryHost=always', async (t) => {
+    const testdir = t.testdir({
+      project: {
+        'package.json': JSON.stringify({
+          name: 'myproject',
+          version: '1.0.0',
+          dependencies: {
+            abbrev: '1.1.1',
+          },
+        }),
+      },
+    })
+
+    tnock(t, 'https://registry.github.com')
+      .get('/abbrev')
+      .reply(200, abbrevPackument2)
+
+    tnock(t, 'https://registry.github.com')
+      .get('/abbrev/-/abbrev-1.1.1.tgz')
+      .reply(200, abbrevTGZ)
+
+    const arb = new Arborist({
+      path: resolve(testdir, 'project'),
+      registry: 'https://registry.github.com',
+      cache: resolve(testdir, 'cache'),
+      replaceRegistryHost: 'always',
+    })
+    await arb.reify()
+  })
+})
